#!/usr/bin/env python3
# coding:utf-8
from .error import ParsecError
from . import Parsec, oneOf
from io import StringIO
from . import skip, skip1


def string(s, case_sensitive=True):
    @Parsec
    def call(st):
        buffer = StringIO()
        segment = s if case_sensitive else s.lower()
        for chr in segment:
            c = st.next()
            if (chr != c and case_sensitive) or (not case_sensitive and chr != c.lower()):
                raise ParsecError(st, "Expect '{0}' but got {1}".format(s, c))
            else:
                buffer.write(c)
        else:
            return buffer.getvalue()

    return call


def textIn(t, case_sensitive=True):
    check = set(x.lower for x in t) if case_sensitive else t
    return oneOf(check)


@Parsec
def space(state):
    c = state.next()
    if c.isspace():
        return c
    raise ParsecError(state, "Expect a space but got {0}".format(c))


@Parsec
def noSpace(state):
    c = state.next()
    if c.isspace():
        raise ParsecError(state, "Expect any no space but got {0}".format(c))
    return c


@Parsec
def digit(state):
    c = state.next()
    if c.isdigit():
        return c
    else:
        raise ParsecError(state, "Expect a space but got {0}".format(c))


@Parsec
def skipSpaces(state):
    return skip(space)(state)


@Parsec
def skip1Spaces(state):
    return skip1(space)(state)


def mkString(txt):
    @Parsec
    def call(state):
        return "".join(txt)

    return call


def join():
    def binder(txt):
        @Parsec
        def call(state):
            return "".join(txt)

        return call

    return binder



#!/usr/bin/env python3
# coding:utf-8

class ParsecError(Exception):
    def __init__(self, state, message):
        self.index = state.index
        self.message = message

    def __str__(self):
        return "Error at {0}: {1}".format(self.index, self.message)


class ParsecEof(ParsecError):
    def __init__(self, state):
        ParsecError.__init__(self, state, "eof")
        # self.index = state.index
        # self.message = "eof"



#!/usr/bin/env python3
# coding:utf-8

class Parsec(object):
    def __init__(self, parsec):
        self.parsec = parsec

    def __call__(self, st):
        return self.parsec(st)

    def bind(self, continuation):
        def bind(st):
            binder = continuation(self.parsec(st))
            return binder(st)

        return Parsec(bind)

    def then(self, p):
        def then(st):
            self.parsec(st)
            return p(st)

        return Parsec(then)

    def over(self, p):
        def over(st):
            re = self.parsec(st)
            p(st)
            return re

        return Parsec(over)



import json
import logging

SUCCESS = 0

log = logging.getLogger(__name__)


class StateError(Exception):
    def __init__(self, error):
        super(error)
        self.error = error


class Result:
    def __init__(self, status, value):
        self.logger = log
        self.status = status
        self.value = value

    @classmethod
    def success(cls, value):
        log.debug(f"create success value of result[{value}]")
        return Result(SUCCESS, value)

    @classmethod
    def failed(cls, error):
        log.debug(f"create failure error of [{error}]")
        return Result(error, None)

    def map(self, func):
        if self.isOk():
            self.logger.debug(f"call map function with value [{self.value}]")
            call = result(func)
            return call(self.value)
        else:
            self.logger.debug(f"ignore map call because result failure [{self.status}]")
            return self

    def flatMap(self, func):
        if self.isOk():
            self.logger.debug(f"call flatmap function with value [{self.value}]")
            return func(self.value)
        else:
            self.logger.debug(f"ignore flatmap call because result failure [{self.status}]")
            return self

    def isOk(self):
        return self.status == SUCCESS

    def isFailed(self):
        return self.status != SUCCESS

    def get(self):
        if self.isOk():
            return self.value
        else:
            raise StateError(self.status)

    def json(self):
        """
        生成 json 对象，要求 value 本身是可以被 json 库处理的数据结构，或者提供了 json 方法。
        :return:生成的 json 字符串。包含状态和数据。
        """
        if self.isOk():
            status = "ok"
        else:
            status = self.status

        if hasattr(self.value, "json"):
            data = self.value.json()
        else:
            data = self.value

        return json.dumps({
            "state": status,
            "result": data
        })


def success(result):
    return Result(SUCCESS, result)


def failed(error):
    return Result(error, None)


def result(func):
    def call(*args, **argv):
        try:
            value = func(*args, **argv)
            return success(value)
        except Exception as error:
            return failed(error)

    return call


def either(func):
    def call(*args, **argv):
        try:
            error, value = func(*args, **argv)
            status = SUCCESS if error is None else error
            return Result(status, value)
        except Exception as error:
            return failed(error)

    return call



#!/usr/bin/python3
# coding:utf-8
from .error import ParsecEof


class BasicState(object):
    def __init__(self, data):
        self.data = data
        self.index = 0
        self.tran = -1

    def next(self):
        if 0 <= self.index < len(self.data):
            re = self.data[self.index]
            self.index += 1
            return re
        else:
            raise ParsecEof(self)

    def begin(self):
        if self.tran == -1:
            self.tran = self.index

        return self.index

    def commit(self, tran):
        if self.tran == tran:
            self.tran = -1

    def rollback(self, tran):
        self.index = tran
        if self.tran == tran:
            self.tran = -1



#!/usr/bin/env python3
# coding:utf-8

import unittest

from parsec import BasicState, one, eof, ParsecEof, ne, eq, oneOf, noneOf, pack, fail, ParsecError

simple = "It is a simple string."


class TestAtom(unittest.TestCase):
    def test_one(self):
        st = BasicState(simple)
        for i in range(len(simple)):
            idx = st.index
            re = one(st)
            self.assertEqual(re, st.data[idx])
        with self.assertRaises(Exception) as err:
            one(st)
        self.assertTrue(issubclass(type(err.exception), ParsecEof))

    def test_eof(self):
        st = BasicState(simple)
        for i in range(len(simple)):
            re = one(st)
        self.assertIsNone(eof(st))

    def test_eq(self):
        st = BasicState(simple)
        for i in range(len(simple)):
            c = st.data[st.index]
            next_parser = eq(c)
            re = next_parser(st)
            self.assertEqual(re, c)

    def test_ne(self):
        st = BasicState(simple)
        next_parser = ne("")
        for i in range(len(simple)):
            c = st.data[st.index]
            re = next_parser(st)
            self.assertEqual(re, c)
            next_parser = ne(c)

    def test_one_of(self):
        st = BasicState(simple)
        next_parser = oneOf(simple)
        for i in range(len(simple)):
            idx = st.index
            re = next_parser(st)
            self.assertEqual(re, st.data[idx])

    def test_none_of(self):
        st = BasicState(simple)
        n = noneOf("xyzcf")
        for i in range(1, len(simple)):
            idx = st.index
            re = n(st)
            self.assertEqual(re, st.data[idx])

    def test_pack(self):
        st = BasicState(simple)
        p = pack("z")
        for i in range(1, len(simple)):
            idx = st.index
            re = p(st)
            self.assertEqual("z", re)

    def test_fail(self):
        st = BasicState(simple)
        p = fail("z")
        with self.assertRaises(ParsecError):
            p(st)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# coding:utf-8

from . import Parsec
from . import ParsecError, ParsecEof


@Parsec
def one(st):
    return st.next()


def eq(data):
    @Parsec
    def call(st):
        re = st.next()
        if re == data:
            return re
        else:
            raise ParsecError(st, "Expect {0} but got {1}".format(data, re))

    return call


def ne(data):
    @Parsec
    def call(st):
        re = st.next()
        if re != data:
            return re
        else:
            raise ParsecError(st, "Expect not equal {0}".format(data))

    return call


def oneOf(data):
    @Parsec
    def call(st):
        re = st.next()
        if re in data:
            return re
        else:
            raise ParsecError(st, "Expect one item of {0} but got {1}".format(data, re))

    return call


def noneOf(data):
    @Parsec
    def call(st):
        re = st.next()
        if re in data:
            raise ParsecError(st, "Expect none item of {0} but got {1}".format(data, re))
        else:
            return re

    return call


def pack(data):
    @Parsec
    def call(st):
        return data

    return call


def fail(message):
    @Parsec
    def call(st):
        raise ParsecError(st, message)

    return call


@Parsec
def eof(state):
    re = None
    try:
        re = state.next()
    except ParsecEof:
        return None
    raise ParsecError(state, "Expect eof but got {0}".format(re))



#!/usr/bin/env python3
# coding:utf-8

# N = [0..9] | (T); F = N | F * N; T = F | T + F; P = T$;
# 输入文本是"1 * 2 + 3 * (4 + 5)
# 因为parsec不支持左递归，这个脚本一定会死循环，所以不要直接执行

import string

from src.parsec import oneOf, Parsec, many1, eq, choice, attempt, between, BasicState

input_expr: str = "1*2+3*(4+5)"

digit = oneOf(string.digits)


@Parsec
def integer(st):
    re = many1(digit)(st)
    return int("".join(re))


def mulN(x):
    @Parsec
    def func(st):
        y = eq("*").then(N)(st)
        return x * y

    return func


def addF(x):
    @Parsec
    def func(st):
        y = eq("+").then(F)(st)
        return x + y

    return func


@Parsec
def N(st):
    return choice(attempt(integer), between(eq('('), eq(')'), T))(st)


@Parsec
def F(st):
    # x = choice(attempt(N), F)(st)
    # eq('*')(st)
    # y = N(st)
    # return x * y
    return choice(attempt(N), F).bind(mulN)(st)


@Parsec
def T(st):
    # x = choices(attempt(F), T)(st)
    # eq('+')(st)
    # y = F(st)
    # return x + y
    return choice(attempt(F), T).bind(addF)(st)


if __name__ == '__main__':
    st = BasicState(input_expr)
    re = T(st)
    print(re)



#!/usr/bin/env python3
# coding:utf-8

import unittest

from src.parsec import many, BasicState, one, eq, many1, ParsecError, sep1, ne, sep

simple = "It is a simple string."


class TestCombinator(unittest.TestCase):
    def test_many_0(self):
        st = BasicState(simple)
        p = many(one)
        re = p(st)
        data = "".join(re)
        self.assertEqual(data, simple)

    def test_many_1(self):
        st = BasicState(simple)
        p = many(eq("I"))
        re = p(st)
        data = "".join(re)
        self.assertEqual(data, "I")

    def test_many_2(self):
        st = BasicState(simple)
        p = many(eq("z"))
        re = p(st)
        data = "".join(re)
        self.assertEqual(data, "")

    def test_many0_1(self):
        st = BasicState(simple)
        p = many1(one)
        re = p(st)
        data = "".join(re)
        self.assertEqual(data, simple)

    def test_many1_1(self):
        st = BasicState(simple)
        p = many1(eq("I"))
        re = p(st)
        data = "".join(re)
        self.assertEqual(data, "I")

    def test_many1_2(self):
        st = BasicState(simple)
        p = many1(eq("z"))
        with self.assertRaises(ParsecError):
            p(st)

    def test_sep_0(self):
        st = BasicState(simple)
        p = sep(many1(ne(" ")), eq(" "), )
        re = p(st)
        data = ["".join(item) for item in re]
        self.assertEqual(data, simple.split(" "))

    def test_sep1_0(self):
        st = BasicState(simple)
        p = sep1(many1(ne(" ")), eq(" "))
        re = p(st)
        data = ["".join(item) for item in re]
        self.assertEqual(data, simple.split(" "))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# coding:utf-8

from .error import *
from .parsec import Parsec
from .state import BasicState
from .atom import one, eof, eq, ne, oneOf, noneOf, pack, fail
from .combinator import attempt, choice, choices, many, many1, manyTill, sep, sep1, sepTail, sep1Tail, skip, skip1, \
    between
from .text import string, space, digit

__all__ = ["Parsec", "BasicState", "one", "eof", "eq", "ne", "oneOf", "noneOf",
           "pack", "fail", "attempt", "choice", "choices", "many", "many1", "manyTill",
           "between", "sep", "sep1", "sepTail", "sep1Tail", "skip", "skip1", "string",
           "space", "digit", "ParsecEof", "ParsecError"]



#!/usr/bin/env python3
# coding:utf-8

from .error import ParsecError
from . import Parsec, pack


def attempt(p):
    @Parsec
    def call(state):
        tran = state.begin()
        try:
            re = p(state)
            state.commit(tran)
            return re
        except Exception as err:
            state.rollback(tran)
            raise err

    return call


class Either(Parsec):
    def __init__(self, x, y):
        def call(state):
            prev = state.index
            try:
                return x(state)
            except:
                if self.index == prev:
                    return y(state)
                else:
                    raise

        Parsec.__init__(call)

    def Or(self, continuation):
        return Either(self, continuation)


def choice(x, y):
    @Parsec
    def call(st):
        prev = st.index
        try:
            return x(st)
        except:
            if st.index == prev:
                return y(st)
            else:
                raise

    return call


def choices(*psc):
    if len(psc) < 2:
        raise Exception("choices need more args than one.")

    @Parsec
    def call(st):
        for p in psc[:-1]:
            prev = st.index
            try:
                return p(st)
            except:
                if st.index != prev:
                    raise
        else:
            return psc[-1](st)

    return call


def many(p):
    return choice(attempt(many1(p)), pack([]))


def many1(p):
    @Parsec
    def call(st):
        re = [p(st)]
        try:
            while True:
                re.append(attempt(p)(st))
        except Exception as err:
            pass
        finally:
            return re

    return call

def skip(p):
    @Parsec
    def call(state):
        while True:
            tran = state.begin()
            try:
                p(state)
                state.commit(tran)
            except ParsecError:
                tran.rollback(tran)
                return
        
    return call

def skip1(p):
    @Parsec
    def call(state):
        p(state)
        skip(p)(state)
        
    return call

def sep(p, s):
    return choice(attempt(sep1(p, s)), pack([]))


def sep1(p, s):
    @Parsec
    def call(state):
        re = [p(state)]
        try:
            while True:
                re.append(attempt(s.then(p))(state))
        except Exception as err:
            pass
        finally:
            return re

    return call


def manyTill(p, t):
    @Parsec
    def call(state):
        re = []
        end = attempt(t)
        while True:
            try:
                end(state)
                return re
            except Exception as err:
                re.append(p(state))

    return call


def sepTail(s, p, t):
    return sep(s, p).over(t)


def sep1Tail(s, p, t):
    return sep1(s, p).over(t)


def manyTill(p, t):
    parser = attempt(p)
    stop = attempt(t)
    @Parsec
    def call(state):
        result = []
        while True:
            try:
                t(state)
                return result
            except ParsecError:
                element = parser(state)
                result.append(element)


def skip(s):
    @Parsec
    def call(state):
        try:
            while True:
                attempt(s)(state)
        except:
            pass
        finally:
            return None

    return call


def skip1(s):
    @Parsec
    def call(state):
        s(state)
        try:
            while True:
                attempt(s)(state)
        except:
            pass
        finally:
            return None

    return call


def between(openp, closep, psc):
    @Parsec
    def call(state):
        openp(state)
        re = psc(state)
        closep(state)
        return re

    return call

def ahead(psc):
    @Parsec
    def call(state):
        tran = state.begin()
        try:
            return psc(state)
        finally:
            state.rollback(tran)


class BuiltIn:
    def __init__(self, parsec):
        self.parsec = parsec

    @property
    def head(self):
        return ahead(self.parsec)

    @property
    def attempt(self):
        return attempt(self.parsec)

    @property
    def many(self):
        return many(self.parsec)        

    @property
    def many1(self):
        return many1(self.parsec)

    @property
    def skip(self):
        return skip(self.parsec)

    @property
    def skip1(self):
        return skip1(self.parsec)

    def sepBy(self, by):
        return sep(self.parsec, by)

    def sepBy1(self, by):
        return sep1(self.parsec, by)

    def otherwise(self, other):
        return choice(self.attempt, other)

    def manyTill(self, end):
        return manyTill(self,parserc, end)


import unittest

from src.parsec import BasicState, string


class TestTextParsers(unittest.TestCase):
    def test_text(self):
        content = "this is a simple string"
        state = BasicState(content)
        p = string("this")
        self.assertEqual(p(state), "this")

    def test_case(self):
        state = BasicState("this is a simple string")
        p = string("THIS", False)
        self.assertEqual(p(state), "this")

        state = BasicState("This is a simple string")
        p = string("this", False)
        self.assertEqual(p(state), "This")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# coding:utf-8

import unittest

from src.parsec import BasicState, ParsecEof

sample = "It is a simple string."


class TestState(unittest.TestCase):
    def test_next(self):
        st = BasicState(sample)
        for i in range(len(sample)):
            idx = st.index
            re = st.next()
            self.assertEqual(re, st.data[idx])
        with self.assertRaises(Exception) as err:
            st.next()
        self.assertTrue(issubclass(type(err.exception), ParsecEof))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from setuptools import setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name="pyparsec",
      version="0.7.1",
      description="Parsec From Haskell Python 3 Portable",
      long_description=long_description,
      author="marsliu",
      author_email="mars.liu@outlook.com",
      url="https://github.com/Dwarfartisan/pyparsec",
      license="MIT",
      packages=["parsec", "test"],
      package_dir={
          "parsec": "src/parsec",
          "test": "src/tests"
      },
      classifiers=[
         "Topic :: Utilities",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3 :: Only",
         "License :: OSI Approved :: MIT License"
      ]
)
